@@ -25,14 +25,25 @@ from .contrib import export_to_mirror
 from .contrib import export_pymongo
 from .contrib import fetch_one  # deprecated
 from .contrib import filesystems as fs
+from .contrib import Collection
+from .contrib import index_files
+from .contrib import index
+from .contrib import RegexFileCrawler
+from .contrib import MasterCrawler
+from .contrib import SignacProjectCrawler
 from .db import get_database
 
-__version__ = '0.7.1'
+__version__ = '0.8.0'
 
 __all__ = ['__version__', 'contrib', 'db', 'errors',
            'cite',
            'Project', 'get_project', 'init_project',
            'get_database', 'fetch', 'fetch_one',
            'export_one', 'export', 'export_to_mirror',
-           'export_pymongo', 'fs'
+           'Collection',
+           'export_pymongo', 'fs',
+           'index_files', 'index',
+           'RegexFileCrawler',
+           'MasterCrawler',
+           'SignacProjectCrawler',
            ]
